"""
This file is part of pys5p

https://github.com/rmvanhees/pys5p.git

The class L1Bpatch provides methods to patch Tropomi SWIR measurement data in
offline level 1b products (incl. calibration, irradiance and radiance).

Copyright (c) 2017-2022 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from datetime import datetime
from pathlib import Path
import shutil

from setuptools_scm import get_version

import h5py
import numpy as np

from .l1b_io import L1BioRAD
from . import swir_region


# - global variables --------------------------------
_MSG_ERR_IO_BAND_ = 'spectral band of input and output products do not match'


# - local functions --------------------------------


# - class definition -------------------------------
class L1Bpatch():
    """
    Definition off class L1Bpatch
    """
    def __init__(self, l1b_product: str, data_dir='/tmp',
                 ckd_dir='/nfs/Tropomi/share/ckd') -> None:
        """
        Initialize access to a Tropomi offline L1b product
        """
        prod_type = Path(l1b_product).name[0:15]
        if prod_type not in ('S5P_OFFL_L1B_RA', 'S5P_RPRO_L1B_RA'):
            raise TypeError(
                'Warning: only implemented for Tropomi L1b radiance products')

        # initialize private class-attributes
        self.data_dir = Path(data_dir)
        if not self.data_dir.is_dir():
            self.data_dir.mkdir(mode=0o755)
        self.ckd_dir = Path(ckd_dir)
        self.l1b_product = Path(l1b_product)
        self.l1b_patched = \
            self.data_dir / self.l1b_product.name.replace('_01_', '_99_')
        if self.l1b_patched.is_file():
            self.l1b_patched.unlink()
        self.__patched_msm = []

    def __enter__(self):
        """
        method called to initiate the context manager
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        """
        method called when exiting the context manager
        """
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self) -> None:
        """
        Before closing the product, we make sure that the output product
        describes what has been altered by the S/W. To keep any change
        traceable.

        In case the L1b product is altered, the attributes listed below are
        added to the group: '/METADATA/SRON_METADATA':
             - dateStamp ('now')
             - Git-version of S/W
             - list of patched datasets
             - auxiliary datasets used by patch-routines
        """
        if not self.l1b_patched.is_file():
            return

        if not self.__patched_msm:
            return

        with h5py.File(self.l1b_patched, 'r+') as fid:
            sgrp = fid.require_group('/METADATA/SRON_METADATA')
            sgrp.attrs['dateStamp'] = datetime.utcnow().isoformat()
            sgrp.attrs['git_tag'] = get_version(root='..',
                                                relative_to=__file__)
            if 'patched_datasets' not in sgrp:
                dtype = h5py.special_dtype(vlen=str)
                dset = sgrp.create_dataset('patched_datasets',
                                           (len(self.__patched_msm),),
                                           maxshape=(None,), dtype=dtype)
                dset[:] = np.asarray(self.__patched_msm)
            else:
                dset = sgrp['patched_datasets']
                dset.resize(dset.shape[0] + len(self.__patched_msm), axis=0)
                dset[dset.shape[0]-1:] = np.asarray(self.__patched_msm)

    # --------------------------------------------------
    def pixel_quality(self, dpqm, threshold=0.8) -> None:
        """
        Patch SWIR pixel_quality.

        Patched dataset: 'quality_level' and 'spectral_channel_quality'

        Requires (naive approach):
        * read original dataset 'spectral_channel_quality'
        * read pixel quality ckd
        * adjust second pixel of each byte of spectral_channel_quality
        * quality_level = int(100 * dpqm)
        * write updated datasets to patched product

        Parameters
        ----------
        dpqm  :   array-like
           SWIR pixel quality as a float value between 0 and 1
        thershold :  float, optional
           threshold for good pixels, default 0.8

        Returns
        -------
        Nothing
        """
        if not self.l1b_patched.is_file():
            shutil.copy(self.l1b_product, self.l1b_patched)

        # read original data
        with L1BioRAD(self.l1b_product) as l1b:
            band = l1b.select('STANDARD_MODE')
            quality_level = l1b.get_msm_data('quality_level')
            print('quality_level', quality_level.dtype)
            chan_quality = l1b.get_msm_data('spectral_channel_quality')
            print('chan_quality', chan_quality.dtype)

        if band in ('7', '8'):
            l2_dpqm = dpqm[swir_region.coords(mode='level2', band=band)]
        else:
            raise ValueError('only implemented for band 7 or 8')

        # patch dataset 'quality_level'
        quality_level[...] = (100 * l2_dpqm).astype(np.uint8)  # broadcasting

        # patch dataset 'spectral_channel_quality'
        buff = chan_quality & ~2          # set second bit to zero (all good)
        buff[:, l2_dpqm < threshold] += 2                   # flag bad pixels
        chan_quality = buff.astype(np.uint8)

        # write patched dataset to new product
        with L1BioRAD(self.l1b_patched, readwrite=True) as l1b:
            res = l1b.select('STANDARD_MODE')
            if res != band:
                raise ValueError(_MSG_ERR_IO_BAND_)
            l1b.set_msm_data('quality_level', quality_level)
            l1b.set_msm_data('spectral_channel_quality', chan_quality)

    def offset(self) -> None:
        """
        Patch SWIR offset correction.

        Patched dataset: 'radiance' ('radiance_error' and 'radiance_noise'?)

        Requires (naive approach):
        * reverse applied radiance calibration
        * reverse applied stray-light correction
        * reverse applied PRNU correction
        * reverse applied dark-flux correction
        * reverse applied offset correction
        * apply (alternative) offset correction
        * apply (alternative) dark-flux correction
        * apply (alternative) PRNU correction
        * apply (alternative) stray-light correction
        * apply (alternative) radiance calibration

        Parameters
        ----------
        None

        Returns
        -------
        Nothing
        """
        if not self.l1b_patched.is_file():
            shutil.copy(self.l1b_product, self.l1b_patched)

        # read original data
        with L1BioRAD(self.l1b_product) as l1b:
            band = l1b.select('STANDARD_MODE')
            data = l1b.get_msm_data('radiance')

        # read required CKD's

        # patch dataset 'radiance'

        # write patched dataset to new product
        with L1BioRAD(self.l1b_patched, readwrite=True) as l1b:
            res = l1b.select('STANDARD_MODE')
            if res != band:
                raise ValueError(_MSG_ERR_IO_BAND_)
            l1b.set_msm_data('radiance', data)

    def darkflux(self) -> None:
        """
        Patch SWIR dark-flux correction.

        Patched dataset: 'radiance' ('radiance_error' and 'radiance_noise'?)

        Requires (naive approach):
        * reverse applied radiance calibration
        * reverse applied stray-light correction
        * reverse applied PRNU correction
        * reverse applied dark-flux correction
        * apply (alternative) dark-flux correction
        * apply (alternative) PRNU correction
        * apply (alternative) stray-light correction
        * apply (alternative) radiance calibration

        Parameters
        ----------
        None

        Returns
        -------
        Nothing
        """
        if not self.l1b_patched.is_file():
            shutil.copy(self.l1b_product, self.l1b_patched)

        # read original data
        with L1BioRAD(self.l1b_product) as l1b:
            band = l1b.select('STANDARD_MODE')
            data = l1b.get_msm_data('radiance')

        # read required CKD's

        # patch dataset 'radiance'

        # write patched dataset to new product
        with L1BioRAD(self.l1b_patched, readwrite=True) as l1b:
            res = l1b.select('STANDARD_MODE')
            if res != band:
                raise ValueError(_MSG_ERR_IO_BAND_)
            l1b.set_msm_data('radiance', data)

    def prnu(self) -> None:
        """
        Patch pixel response non-uniformity correction.

        Patched dataset: 'radiance' ('radiance_error' and 'radiance_noise'?)

        Requires (naive approach):
        * reverse applied radiance calibration
        * reverse applied stray-light correction
        * reverse applied PRNU correction
        * apply (alternative) PRNU correction
        * apply (alternative) stray-light correction
        * apply (alternative) radiance calibration

        Alternative: neglect impact stray-light, but apply patch to correct for
        spectral features.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Notes
        -----
        It is assumed that for the PRNU correction the CKD has to be multiplied
        with the pixel signals.
        """
        if not self.l1b_patched.is_file():
            shutil.copy(self.l1b_product, self.l1b_patched)

        # read original data
        with L1BioRAD(self.l1b_product) as l1b:
            band = l1b.select('STANDARD_MODE')
            data = l1b.get_msm_data('radiance')

        # read required CKD's

        # patch dataset 'radiance'

        # write patched dataset to new product
        with L1BioRAD(self.l1b_patched, readwrite=True) as l1b:
            res = l1b.select('STANDARD_MODE')
            if res != band:
                raise ValueError(_MSG_ERR_IO_BAND_)
            l1b.set_msm_data('radiance', data)

    def relrad(self) -> None:
        """
        Patch relative radiance calibration.

        Patched dataset: 'radiance' ('radiance_error' and 'radiance_noise'?)

        Requires:
        * reverse applied radiance calibration
        * apply alternative radiance calibration

        Parameters
        ----------
        None

        Returns
        -------
        Nothing
        """
        if not self.l1b_patched.is_file():
            shutil.copy(self.l1b_product, self.l1b_patched)

        # read original data
        with L1BioRAD(self.l1b_product) as l1b:
            band = l1b.select('STANDARD_MODE')
            data = l1b.get_msm_data('radiance')

        # read required CKD's

        # patch dataset 'radiance'

        # write patched dataset to new product
        with L1BioRAD(self.l1b_patched, readwrite=True) as l1b:
            res = l1b.select('STANDARD_MODE')
            if res != band:
                raise ValueError(_MSG_ERR_IO_BAND_)
            l1b.set_msm_data('radiance', data)

    def absrad(self) -> None:
        """
        Patch absolute radiance calibration.

        Patched dataset: 'radiance' ('radiance_error' and 'radiance_noise'?)

        Requires:
        * reverse applied irradiance calibration
        * apply alternative irradiance calibration

        Parameters
        ----------
        None

        Returns
        -------
        Nothing
        """
        if not self.l1b_patched.is_file():
            shutil.copy(self.l1b_product, self.l1b_patched)

        # read original data
        with L1BioRAD(self.l1b_product) as l1b:
            band = l1b.select('STANDARD_MODE')
            data = l1b.get_msm_data('radiance')

        # read required CKD's

        # patch dataset 'radiance'

        # write patched dataset to new product
        with L1BioRAD(self.l1b_patched, readwrite=True) as l1b:
            res = l1b.select('STANDARD_MODE')
            if res != band:
                raise ValueError(_MSG_ERR_IO_BAND_)
            l1b.set_msm_data('radiance', data)

    def check(self) -> None:
        """
        Check patched dataset in L1B product
        """
        if not self.l1b_patched.is_file():
            raise ValueError('patched product not found')

        with h5py.File(self.l1b_patched, 'r+') as fid:
            if 'SRON_METADATA' not in fid['/METADATA']:
                raise ValueError('no SRON metadata defined in L1B product')
            sgrp = fid['/METADATA/SRON_METADATA']
            if 'patched_datasets' not in sgrp:
                raise ValueError('no patched datasets in L1B prduct')
            patched_datasets = sgrp['patched_datasets'][:]

        for ds_name in patched_datasets:
            with L1BioRAD(self.l1b_product) as l1b:
                l1b.select('STANDARD_MODE')
                orig = l1b.get_msm_data(ds_name.split('/')[-1])

            with L1BioRAD(self.l1b_patched) as l1b:
                l1b.select('STANDARD_MODE')
                patch = l1b.get_msm_data(ds_name.split('/')[-1])

            if np.issubdtype(orig.dtype, np.integer):
                if np.array_equiv(orig, patch):
                    print(ds_name.split('/')[-1], ' equal True')
                else:
                    print(f'{ds_name.split("/")[-1]}'
                          f' equal {(orig == patch).sum()}'
                          f' differ {(orig != patch).sum()}')
            else:
                print('test not yet defined')



"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

Performs unit-tests on class CKDio (xarray version)

Copyright (c) 2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import argparse
from pathlib import Path

import numpy as np

from pys5p.lv2_io import LV2io

def read_lv2(l2_product):
    """
    Read Tropomi level 2 product
    """
    with LV2io(l2_product) as lv2:
        # Class properties
        print('science_product: ', lv2.science_product)
        print('orbit: ', lv2.orbit)
        print('algorithm_version: ', lv2.algorithm_version)
        print('processor_version: ', lv2.processor_version)
        print('product_version: ', lv2.product_version)
        if not lv2.science_product:
            print('coverage_time: ', lv2.coverage_time)
        print('creation_time: ', lv2.creation_time)
        # Attributes
        print('get_attr: ', lv2.get_attr('title'))
        if lv2.science_product:
            ds_name = 'h2o_column'
        else:
            ds_name = 'methane_mixing_ratio'
        print('get_attr: ', lv2.get_attr('long_name', ds_name))

        # Time information
        print('ref_time: ', lv2.ref_time)
        print('get_time: ', lv2.get_time())
        # Geolocation
        geo_var = 'latitude_center' if lv2.science_product else 'latitude'
        print('get_geo_data: ', lv2.get_geo_data()[geo_var].shape)
        # Footprints
        geo_var = 'latitude'
        print('get_geo_bounds: ', lv2.get_geo_bounds()[geo_var].shape)
        print('get_geo_bounds: ', lv2.get_geo_bounds(
            data_sel=np.s_[250:300, 100:110])[geo_var].shape)
        # Datasets (numpy)
        if lv2.science_product:
            ds_name = 'h2o_column'
        else:
            ds_name = 'methane_mixing_ratio'
        print('get_dataset: ', lv2.get_dataset(ds_name).shape)
        print('get_dataset: ', lv2.get_dataset(
            ds_name, data_sel=np.s_[250:300, 100:110]).shape)
        # Datasets (xarray)
        if lv2.science_product:
            ds_name = 'h2o_column'
        else:
            ds_name = 'methane_mixing_ratio'
        print('get_data_as_xds: ', lv2.get_data_as_xds(ds_name))


def main():
    """
    Main function
    """
    # parse command-line parameters
    parser = argparse.ArgumentParser(
        description=f'{Path(__file__).name}: run units-test on class LV2io')
    parser.add_argument('lv2_product', nargs=1, type=str, default=None,
                        help='use this Tropomi level2 product')
    args = parser.parse_args()
    print(args)

    read_lv2(args.lv2_product[0])


# - main code --------------------------------------
if __name__ == '__main__':
    main()



"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

There are two definitions of the usable area on the SWIR detector:
 - 'illuminated' - detector area illuminated by external sources,
 defined as an rectangular area where the signal is at least 50% of the
 maximum signal. Coordinates: rows [11:228], columns [16:991].
 - 'level2' - a smaller area used in official SWIR level 1B (ir)radiance
 products. Coordinates: rows [12:227], columns [20:980].

Note row 257 of the SWIR detector is neglected

Copyright (c) 2017-2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import numpy as np


def coords(mode='illuminated', band='78') -> slice:
    """
    Return slice defining the illuminated region on the SWIR detector

    Parameters
    ----------
    mode  :   {'illuminated', 'level2'}, optional
       default is 'illuminated'
    band  :   str, optional
       select band 7 or 8, default is both bands
    """
    if mode == 'level2':
        if band == '7':
            return np.s_[12:227, 20:500]
        if band == '8':
            return np.s_[12:227, :480]
        # else
        return np.s_[12:227, 20:980]

    if band == '7':
        return np.s_[11:228, 16:500]
    if band == '8':
        return np.s_[11:228, :491]
    # else
    return np.s_[11:228, 16:991]


def mask(mode='illuminated', band='78'):
    """
    Return mask of the illuminated region, where the value of the illuminated
    pixels are set to True.

    Parameters
    ----------
    mode  :   {'illuminated', 'level2'}, optional
       default is 'illuminated'
    band  :   str, optional
       select band 7 or 8, default is both bands
    """
    if band in ('7', '8'):
        res = np.full((256, 500), False)
    else:
        res = np.full((256, 1000), False)

    res[coords(mode, band)] = True

    return res



"""
This file is part of pys5p

https://github.com/rmvanhees/pys5p.git

Calculates the exact pixel exposure time of the SWIR measurements

Requires parameters int_delay and int_holt from the instrument_settings

Copyright (c) 2019-2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""


def swir_exp_time(int_delay: int, int_hold: int) -> float:
    """
    Parameters
    ----------
    int_delay :  int
      parameters int_delay from the instrument_settings
    int_hold  :  int
      parameters int_holt from the instrument_settings
    Returns
    -------
    float
        exact (SWIR) pixel exposure time
    """
    return 1.25e-6 * (65540 - int_delay + int_hold)



"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

Performs unit-tests on class CKDio (xarray version)

Copyright (c) 2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import argparse
from pathlib import Path

from pys5p.ckd_io import CKDio


def main():
    """
    main function
    """
    parser = argparse.ArgumentParser(
        description=f'{Path(__file__).name}: run units-test on class CKDio')
    parser.add_argument('ckd_dir', nargs=1, type=str, default=None,
                        help=('directory with CKD data with'
                              ' static CKD in a subdirectory static'))
    args = parser.parse_args()

    with CKDio(args.ckd_dir[0], ckd_version=1) as ckd:
        print(ckd.ckd_file)
        for meth in dir(ckd):
            if (meth.startswith('_')
                or meth.startswith('ckd')
                or meth in ('close', 'fid', 'get_param')):
                continue
            print('-------------------------', meth, '[v1]',
                  '-------------------------')
            print(meth, getattr(ckd, meth)())

    with CKDio(args.ckd_dir[0], ckd_version=2) as ckd:
        print(ckd.ckd_file)
        for meth in dir(ckd):
            if (meth.startswith('_')
                or meth.startswith('ckd')
                or meth in ('close', 'fid', 'get_param')):
                continue
            print('-------------------------', meth, '[v2]',
                  '-------------------------')
            print(meth, getattr(ckd, meth)())

# - main code --------------------------------------
if __name__ == '__main__':
    main()



"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

Provides access to the S5P Tropomi CKD (static and dynamic)

ToDo
----
 - access to UVN CKD, still incomplete
 - identify latest Static CKD product, e.g. using the validity period

Copyright (c) 2018-2022 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from pathlib import Path, PosixPath

import h5py
import numpy as np
import xarray as xr

from moniplot.image_to_xarray import h5_to_xr


# - local functions ------------------------------
def reject_row257(xarr):
    """
    Remove row 257 from DataArray or Dataset
    """
    return xarr.isel(row=np.s_[0:256])


# - class definition -------------------------------
class CKDio():
    """
    Read Tropomi CKD from the Static CKD product or from dynamic CKD products

    Attributes
    ----------
    ckd_dir : pathlib.Path
    ckd_version : int
    ckd_file : pathlib.Path
    ckd_dyn_file : pathlib.Path
    fid : h5py.File

    Methods
    -------
    close()
       Close resources.
    creation_time()
       Returns datetime when the L1b product was created.
    creator_version()
       Returns version of Tropomi L01B processor used to generate this procuct.
    validity_period()
       Return validity period of CKD product as a tuple of 2 datetime objects.
    get_param(ds_name, band='7')
       Returns value(s) of a CKD parameter from the Static CKD product.
    dn2v_factors()
       Returns digital number to Volt CKD, SWIR only.
    v2c_factors()
       Returns Voltage to Charge CKD, SWIR only.
    absirr(qvd=1, bands='78')
       Returns absolute irradiance responsivity.
    absrad(bands='78')
       Returns absolute radiance responsivity.
    memory()
       Returns memory CKD, SWIR only.
    noise()
       Returns pixel read-noise CKD, SWIR only
    prnu(bands='78')
       Returns Pixel Response Non-Uniformity (PRNU).
    relirr(qvd=1, bands='78')
       Returns relative irradiance correction.
    saa()
       Returns definition of the SAA region.
    wavelength(bands='78')
       Returns wavelength CKD.
    darkflux()
       Returns dark-flux CKD, SWIR only.
    offset()
       Returns offset CKD, SWIR only.
    pixel_quality()
       Returns Detector Pixel Quality Mask (float [0, 1]), SWIR only.
    dpqf(threshold=None)
       Returns Detector Pixel Quality Mask (boolean), SWIR only.
    saturation()
       Returns saturation values (pre-offset), SWIR only

    Notes
    -----
    Not all CKD are defined or derived for all bands.
    You can request a CKD for one band or for a channel (bands: '12', '34',
    '56', '78'). Do not mix bands from different channels

    The option to have dynamic CKD is not used for the Tropomi mission, only
    for S/W version 1 a dynamic CKD product is defined. This product contained
    the OCAL CKD and was not updated automatically. For version 2, all CKD are
    stored in one product, where some CKD have a time-axis to correct any
    in-flight degradation.

    Therefore, the logic to find a CKD is implemented as follows:
    1) ckd_dir, defines the base directory to search for the CKD products
    (see below)
    2) ckd_file, defines the full path to (static) CKD product;
    (version 1) any product with dynamic CKD has to be in the same directory

    Version 1:
    * Static CKD are stored in one file: glob('*_AUX_L1_CKD_*')
    * Dynamic CKD are stored in two files:
      - UVN, use glob('*_ICM_CKDUVN_*')
      - SWIR, use glob('*_ICM_CKDSIR_*')

    Version 2+:
    * All CKD in one file: glob('*_AUX_L1_CKD_*')
    * Dynamic CKD are empty

    Examples
    --------
    """
    def __init__(self, ckd_dir=None, ckd_version=1, ckd_file=None):
        """
        Initialize access to a Tropomi Static CKD product

        Parameters
        ----------
        ckd_dir :  str, optional
           Directory where the CKD files are stored,
           default='/nfs/Tropomi/share/ckd'
        ckd_version :  int, optional
           Version of the CKD, default=1
        ckd_file : str, optional
           Name of the CKD file, default=None then the CKD file is searched
           in the directory ckd_dir with ckd_version in the glob-string
        """
        if ckd_dir is None:
            ckd_dir = '/nfs/Tropomi/share/ckd'
        self.ckd_version = max(1, ckd_version)
        self.ckd_dyn_file = None

        # define path to CKD product
        if ckd_file is None:
            if not Path(ckd_dir).is_dir():
                raise FileNotFoundError(f'Not found CKD directory: {ckd_dir}')
            self.ckd_dir = Path(ckd_dir)
            glob_str = f'*_AUX_L1_CKD_*_*_00000_{self.ckd_version:02d}_*_*.h5'
            if (self.ckd_dir / 'static').is_dir():
                res = sorted((self.ckd_dir / 'static').glob(glob_str))
            else:
                res = sorted(self.ckd_dir.glob(glob_str))
            if not res:
                raise FileNotFoundError('Static CKD product not found')
            self.ckd_file = res[-1]
        else:
            if not Path(ckd_file).is_file():
                raise FileNotFoundError(f'Not found CKD file: {ckd_file}')
            self.ckd_dir = Path(ckd_file).parent
            self.ckd_file = Path(ckd_file)

        # obtain path to dynamic CKD product (version 1, only)
        if self.ckd_version == 1:
            if (self.ckd_dir / 'dynamic').is_dir():
                res = sorted((self.ckd_dir / 'dynamic').glob('*_ICM_CKDSIR_*'))
            else:
                res = sorted(self.ckd_dir.glob('*_ICM_CKDSIR_*'))
            if res:
                self.ckd_dyn_file = res[-1]

        # open access to CKD product
        self.fid = h5py.File(self.ckd_file, "r")

    def __enter__(self):
        """
        method called to initiate the context manager
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        method called when exiting the context manager
        """
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self) -> None:
        """
        Make sure that we close all resources
        """
        if self.fid is not None:
            self.fid.close()

    def creation_time(self) -> str:
        """
        Returns datetime when the L1b product was created
        """
        if self.ckd_version == 2:
            attr = self.fid['METADATA'].attrs['production_datetime']
        else:
            group = PosixPath('METADATA', 'earth_explorer_header',
                              'fixed_header', 'source')
            attr = self.fid[str(group)].attrs['Creator_Date'][0]

        if isinstance(attr, bytes):
            attr = attr.decode('ascii')
        return attr

    def creator_version(self) -> str:
        """
        Returns version of Tropomi L01B processor used to generate this procuct
        """
        group = PosixPath('METADATA', 'earth_explorer_header', 'fixed_header')
        attr = self.fid[str(group)].attrs['File_Version']
        if self.ckd_version == 1:
            attr = attr[0]
        if isinstance(attr, bytes):
            attr = attr.decode('ascii')
        return attr

    @staticmethod
    def __get_spectral_channel(bands: str):
        """
        Check bands is valid: single band or belong to one channel

        Parameters
        ----------
        bands : str
           Tropomi bands [1..8] or channels ['12', '34', '56', '78'],
           default: '78'
        """
        band2channel = ['UNKNOWN', 'UV', 'UV', 'VIS', 'VIS',
                        'NIR', 'NIR', 'SWIR', 'SWIR']

        if 0 < len(bands) > 2:
            raise ValueError('read per band or channel, only')

        if len(bands) == 2:
            if band2channel[int(bands[0])] != band2channel[int(bands[1])]:
                raise ValueError('bands should be of the same channel')

        return band2channel[int(bands[0])]

    def get_param(self, ds_name: str, band='7'):
        """
        Returns value(s) of a CKD parameter from the Static CKD product.

        Parameters
        ----------
        ds_name  :  string
           Name of the HDF5 dataset, default='pixel_full_well'
        band     : string
           Band identifier '1', '2', ..., '8', default='7'

        Returns
        -------
        numpy.ndarray or scalar

        Notes
        -----
        Datasets of size=1 are return as scalar

        Handy function for scalar HDF5 datasets, such as:
         - dc_reference_temp
         - dpqf_threshold
         - pixel_full_well
         - pixel_fw_flag_thresh
        """
        if not 1 <= int(band) <= 8:
            raise ValueError('band must be between and 1 and 8')

        if ds_name not in self.fid[f'/BAND{band}']:
            raise ValueError('dataset not available')

        return self.fid[f'/BAND{band}/{ds_name}'][()]

    # ---------- band or channel CKD's ----------
    def dn2v_factors(self):
        """
        Returns digital number to Volt CKD, SWIR only

        Notes
        -----
        The DN2V factor has no error attached to it.
        """
        return np.concatenate(
            (self.fid['/BAND7/dn2v_factor_swir'][2:],
             self.fid['/BAND8/dn2v_factor_swir'][2:]))

    def v2c_factors(self):
        """
        Returns Voltage to Charge CKD, SWIR only

        Notes
        -----
        The V2C factor has no error attached to it.
        """
        # pylint: disable=no-member
        return np.concatenate(
            (self.fid['/BAND7/v2c_factor_swir'].fields('value')[2:],
             self.fid['/BAND8/v2c_factor_swir'].fields('value')[2:]))

    # ---------- spectral-channel CKD's ----------
    def __rd_dataset(self, dset_name: str, bands: str):
        """
        General function to read non-compound dataset into xarray::Dataset

        Parameters
        ----------
        dset_name: str
           name (including path) of the dataset as '/BAND{}/<name>'
        bands : str
           Tropomi bands [1..8] or channels ['12', '34', '56', '78'],
           default: '78'
        """
        ckd_val = None
        for band in bands:
            # try Static-CKD product
            if dset_name.format(band) in self.fid:
                if ckd_val is None:
                    ckd_val = h5_to_xr(self.fid[dset_name.format(band)])
                else:
                    ckd_val = xr.concat(
                        (ckd_val,
                         h5_to_xr(self.fid[dset_name.format(band)])),
                        dim='column')
            # try Dynamic-CKD product
            else:
                dyn_fid = h5py.File(self.ckd_dyn_file, 'r')
                if dset_name.format(band) in dyn_fid:
                    if ckd_val is None:
                        ckd_val = h5_to_xr(dyn_fid[dset_name.format(band)])
                    else:
                        ckd_val = xr.concat(
                            (ckd_val,
                             h5_to_xr(dyn_fid[dset_name.format(band)])),
                            dim='column')
                dyn_fid.close()

        if ckd_val is None:
            return None

        # Use NaN as FillValue
        ckd_val = ckd_val.where(ckd_val != float.fromhex('0x1.ep+122'),
                                other=np.nan)

        # combine DataArrays to Dataset
        return xr.Dataset({'value': ckd_val}, attrs=ckd_val.attrs)

    def __rd_datapoints(self, dset_name: str, bands: str):
        """
        General function to read datapoint dataset into xarray::Dataset

        Parameters
        ----------
        dset_name: str
           name (including path) of the dataset as '/BAND{}/<name>'
        bands : str
           Tropomi bands [1..8] or channels ['12', '34', '56', '78'],
           default: '78'
        """
        ckd_val = None
        ckd_err = None
        for band in bands:
            # try Static-CKD product
            if dset_name.format(band) in self.fid:
                if ckd_val is None:
                    ckd_val = h5_to_xr(self.fid[dset_name.format(band)],
                                       field='value')
                    ckd_err = h5_to_xr(self.fid[dset_name.format(band)],
                                       field='error')
                else:
                    ckd_val = xr.concat(
                        (ckd_val, h5_to_xr(self.fid[dset_name.format(band)],
                                           field='value')), dim='column')
                    ckd_err = xr.concat(
                        (ckd_err, h5_to_xr(self.fid[dset_name.format(band)],
                                           field='error')), dim='column')
            # try Dynamic-CKD product
            else:
                dyn_fid = h5py.File(self.ckd_dyn_file, 'r')
                if dset_name.format(band) in dyn_fid:
                    if ckd_val is None:
                        ckd_val = h5_to_xr(dyn_fid[dset_name.format(band)],
                                           field='value')
                        ckd_err = h5_to_xr(dyn_fid[dset_name.format(band)],
                                           field='error')
                    else:
                        ckd_val = xr.concat(
                            (ckd_val, h5_to_xr(dyn_fid[dset_name.format(band)],
                                               field='value')), dim='column')
                        ckd_err = xr.concat(
                            (ckd_err, h5_to_xr(dyn_fid[dset_name.format(band)],
                                               field='error')), dim='column')
                dyn_fid.close()

        if ckd_val is None:
            return None

        # Use NaN as FillValue
        ckd_val = ckd_val.where(ckd_val != float.fromhex('0x1.ep+122'),
                                other=np.nan)
        ckd_err = ckd_err.where(ckd_err != float.fromhex('0x1.ep+122'),
                                other=np.nan)

        # combine DataArrays to Dataset
        return xr.Dataset({'value': ckd_val, 'error': ckd_err},
                          attrs=ckd_val.attrs)

    # ---------- static CKD's ----------
    def absirr(self, qvd=1, bands='78'):
        """
        Returns absolute irradiance responsivity

        Parameters
        ----------
        bands : str
          Tropomi bands [1..8] or channels ['12', '34', '56', '78'],
          default: '78'
        qvd : int
          Tropomi QVD identifier. Valid values are 1 or 2, default: 1
        """
        try:
            channel = self.__get_spectral_channel(bands)
        except Exception as exc:
            raise RuntimeError(exc) from exc

        dset_name = '/BAND{}' + f'/abs_irr_conv_factor_qvd{qvd}'
        ckd = self.__rd_datapoints(dset_name, bands)
        if '7' in bands or '8' in bands:
            ckd = reject_row257(ckd)
        ckd.attrs["long_name"] = \
                            f'{channel} absolute irradiance CKD (QVD={qvd})'

        return ckd.assign_coords(column=np.arange(ckd.column.size, dtype='u4'))

    def absrad(self, bands='78'):
        """
        Returns absolute radiance responsivity

        Parameters
        ----------
        bands : str
          Tropomi bands [1..8] or channels ['12', '34', '56', '78'],
          default: '78'
        """
        try:
            channel = self.__get_spectral_channel(bands)
        except Exception as exc:
            raise RuntimeError(exc) from exc

        dset_name = '/BAND{}/abs_rad_conv_factor'
        ckd = self.__rd_datapoints(dset_name, bands)
        if '7' in bands or '8' in bands:
            ckd = reject_row257(ckd)
        ckd.attrs["long_name"] = f'{channel} absolute radiance CKD'

        return ckd.assign_coords(column=np.arange(ckd.column.size, dtype='u4'))

    def memory(self):
        """
        Returns memory CKD, SWIR only

        Parameters
        ----------
        bands : str
          Tropomi bands [7,8] or channels ['78'], default: '78'
        """
        column = None
        ckd_parms = ['mem_lin_neg_swir', 'mem_lin_pos_swir',
                     'mem_qua_neg_swir', 'mem_qua_pos_swir']

        ckd = xr.Dataset()
        ckd.attrs["long_name"] = 'SWIR memory CKD'
        for key in ckd_parms:
            dset_name = f'/BAND7/{key}'
            ckd_val = h5_to_xr(self.fid[dset_name], field='value')
            ckd_err = h5_to_xr(self.fid[dset_name], field='error')
            dset_name = f'/BAND8/{key}'
            ckd_val = xr.concat(
                (ckd_val, h5_to_xr(self.fid[dset_name], field='value')),
                dim='column')
            if column is None:
                column = np.arange(ckd_val.column.size, dtype='u4')
            ckd_val = ckd_val.assign_coords(column=column)
            ckd_err = xr.concat(
                (ckd_err, h5_to_xr(self.fid[dset_name], field='error')),
                dim='column')
            ckd_err = ckd_err.assign_coords(column=column)
            ckd[key.replace('swir', 'value')] = reject_row257(ckd_val)
            ckd[key.replace('swir', 'error')] = reject_row257(ckd_err)

        return ckd

    def noise(self):
        """
        Returns readout-noise CKD, SWIR only
        """
        dset_name = '/BAND{}/readout_noise_swir'
        ckd = reject_row257(self.__rd_dataset(dset_name, '78'))
        ckd.attrs["long_name"] = 'SWIR readout-noise CKD'

        return ckd.assign_coords(column=np.arange(ckd.column.size, dtype='u4'))

    def prnu(self, bands='78'):
        """
        Returns Pixel Response Non-Uniformity (PRNU)

        Parameters
        ----------
        bands : str
          Tropomi bands [1..8] or channels ['12', '34', '56', '78'],
          default: '78'
        """
        try:
            channel = self.__get_spectral_channel(bands)
        except Exception as exc:
            raise RuntimeError(exc) from exc

        ckd = self.__rd_datapoints('/BAND{}/PRNU', bands)
        if '7' in bands or '8' in bands:
            ckd = reject_row257(ckd)
        ckd.attrs["long_name"] = f'{channel} PRNU CKD'

        return ckd.assign_coords(column=np.arange(ckd.column.size, dtype='u4'))

    def relirr(self, qvd=1, bands='78'):
        """
        Returns relative irradiance correction

        Parameters
        ----------
        bands : str
          Tropomi bands [1..8] or channels ['12', '34', '56', '78'],
          default: '78'
        qvd : int
          Tropomi QVD identifier. Valid values are 1 or 2, default: 1

        Returns
        -------
        dictionary with keys:
         - band           :    Tropomi spectral band ID
         - mapping_cols   :    coarse irregular mapping of the columns
         - mapping_rows   :    coarse irregular mapping of the rows
         - cheb_coefs     :    chebyshev parameters for elevation and azimuth
                               for pixels on a coarse irregular grid
        """
        try:
            _ = self.__get_spectral_channel(bands)
        except Exception as exc:
            raise RuntimeError(exc) from exc

        res = ()
        for band in bands:
            ckd = {}
            ckd['band'] = int(band)

            dsname = f'/BAND{band}/rel_irr_coarse_mapping_vert'
            ckd['mapping_rows'] = self.fid[dsname][:].astype(int)

            dsname = f'/BAND{band}/rel_irr_coarse_mapping_hor'
            # pylint: disable=no-member
            mapping_hor = self.fid[dsname][:].astype(int)
            mapping_hor[mapping_hor > 1000] -= 2**16
            ckd['mapping_cols'] = mapping_hor

            dsname = f'/BAND{band}/rel_irr_coarse_func_cheb_qvd{qvd}'
            ckd['cheb_coefs'] = self.fid[dsname]['coefs'][:]
            res += (ckd,)

        return res

    def saa(self) -> dict:
        """
        Returns definition of the SAA region
        """
        saa_region = {}
        saa_region['lat'] = self.fid['saa_latitude'][:]
        saa_region['lon'] = self.fid['saa_longitude'][:]

        return saa_region

    def wavelength(self, bands='78'):
        """
        Returns wavelength CKD

        Parameters
        ----------
        bands : str
          Tropomi bands [1..8] or channels ['12', '34', '56', '78'],
          default: '78'

        Notes
        -----
        * The wavelength CKD has no error attached to it.
        """
        try:
            channel = self.__get_spectral_channel(bands)
        except Exception as exc:
            raise RuntimeError(exc) from exc

        dset_name = '/BAND{}/wavelength_map'
        ckd = self.__rd_datapoints(dset_name, bands)
        if '7' in bands or '8' in bands:
            ckd = reject_row257(ckd)
        ckd.attrs["long_name"] = f'{channel} wavelength CKD'

        return ckd.assign_coords(column=np.arange(ckd.column.size, dtype='u4'))

    # ---------- static or dynamic CKD's ----------
    def darkflux(self):
        """
        Returns dark-flux CKD, SWIR only
        """
        dset_name = '/BAND{}/long_term_swir'
        ckd = reject_row257(self.__rd_datapoints(dset_name, '78'))
        ckd.attrs["long_name"] = 'SWIR dark-flux CKD'

        return ckd.assign_coords(column=np.arange(ckd.column.size, dtype='u4'))

    def offset(self):
        """
        Returns offset CKD, SWIR only
        """
        dset_name = '/BAND{}/analog_offset_swir'
        ckd = reject_row257(self.__rd_datapoints(dset_name, '78'))
        ckd.attrs["long_name"] = 'SWIR offset CKD'

        return ckd.assign_coords(column=np.arange(ckd.column.size, dtype='u4'))

    def pixel_quality(self):
        """
        Returns detector pixel-quality mask (float [0, 1]), SWIR only
        """
        dset_name = '/BAND{}/dpqf_map'
        ckd = reject_row257(self.__rd_dataset(dset_name, '78'))
        ckd.attrs["long_name"] = 'SWIR pixel-quality CKD'

        return ckd.assign_coords(column=np.arange(ckd.column.size, dtype='u4'))

    def dpqf(self, threshold=None):
        """
        Returns detector pixel-quality flags (boolean), SWIR only

        Parameters
        ----------
        threshold: float
          Value between [0..1]

        Returns
        -------
        numpy ndarray
        """
        if threshold is None:
            threshold = self.fid['/BAND7/dpqf_threshold'][:]

        # try Static-CKD product
        if '/BAND7/dpqf_map' in self.fid:
            dset = self.fid['/BAND7/dpqf_map']
            dpqf_b7 = dset[:-1, :]
            dset = self.fid['/BAND8/dpqf_map']
            dpqf_b8 = dset[:-1, :]
        else:
            # try Dynamic-CKD product
            with h5py.File(self.ckd_dyn_file, 'r') as fid:
                dset = fid['/BAND7/dpqf_map']
                dpqf_b7 = dset[:-1, :]
                dset = fid['/BAND8/dpqf_map']
                dpqf_b8 = dset[:-1, :]

        return np.hstack((dpqf_b7, dpqf_b8)) < threshold

    def saturation(self):
        """
        Returns pixel-saturation values (pre-offset), SWIR only
        """
        ckd_val = None
        dset_name = '/BAND{}/saturation_preoffset'
        ckd_file = (self.ckd_dir / 'OCAL'
                    / 'ckd.saturation_preoffset.detector4.nc')
        with h5py.File(ckd_file, 'r') as fid:
            ckd_val = xr.concat((h5_to_xr(fid[dset_name.format(7)]),
                                 h5_to_xr(fid[dset_name.format(8)])),
                                dim='column')

        ckd = xr.Dataset({'value': ckd_val}, attrs=ckd_val.attrs)
        ckd = reject_row257(ckd)
        ckd.attrs["long_name"] = 'SWIR pixel-saturation CKD (pre-offset)'

        return ckd.assign_coords(column=np.arange(ckd.column.size, dtype='u4'))



"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

The classe OCMio provide read access to Tropomi on-ground calibration products
(Lx)

Copyright (c) 2017-2022 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from datetime import datetime, timedelta
from pathlib import Path, PurePosixPath

import h5py
import numpy as np

from moniplot.biweight import biweight

# - global parameters ------------------------------


# - local functions --------------------------------
def band2channel(dict_a, dict_b, mode=None):
    """
    Store data from a dictionary as returned by get_msm_data to a ndarray

    Parameters
    ----------
    dict_a      :  dictionary
    dict_b      :  dictionary
    mode        :  list ['combined', 'mean', 'median', 'biweight']
        'combined' will combine data using np.concatenate((data_a, data_b),
                                                          axis=data_a.ndim-1)
        'mean' is calculated using np.nanmean(data, axis=0)
        'median' is calculated using np.nanmedian(data, axis=0)
        'biweight' is calculated using pys5p.biweight(data, axis=0)
        default is None

    Returns
    -------
    out  :  ndarray
        Data from dictionary stored in a numpy array

    Notes
    -----

    Examples
    --------
    >>> data = ocm.band2channel(dict_a, dict_b, mode=['combined', 'median'])
    >>>
    """
    if mode is None:
        mode = []

    data_a = None
    for key in sorted(dict_a):
        buff = dict_a[key][...]

        if data_a is None:
            data_a = buff
        else:
            data_a = np.vstack((data_a, buff))

    if data_a is not None:
        if 'mean' in mode:
            data_a = np.nanmean(data_a, axis=0)
        elif 'median' in mode:
            data_a = np.nanmedian(data_a, axis=0)
        elif 'biweight' in mode:
            data_a = biweight(data_a, axis=0)

    if dict_b is None:
        return data_a

    data_b = None
    for key in sorted(dict_b):
        buff = dict_b[key][...]

        if data_b is None:
            data_b = buff
        else:
            data_b = np.vstack((data_b, buff))

    if data_b is not None:
        if 'mean' in mode:
            data_b = np.nanmean(data_b, axis=0)
        elif 'median' in mode:
            data_b = np.nanmedian(data_b, axis=0)
        elif 'biweight' in mode:
            data_b = biweight(data_b, axis=0)

    if 'combined' in mode:
        return np.concatenate((data_a, data_b), axis=data_a.ndim-1)

    return (data_a, data_b)


# - class definition -------------------------------
class OCMio():
    """
    This class should offer all the necessary functionality to read Tropomi
    on-ground calibration products (Lx)

    Attributes
    ----------
    fid : h5py.File
    filename : string
    band : string

    Methods
    -------
    close()
       Close resources.
    get_processor_version()
       Returns version of the L01b processor used to generate this product.
    get_coverage_time()
       Returns start and end of the measurement coverage time.
    get_attr(attr_name)
       Obtain value of an HDF5 file attribute.
    get_ref_time()
       Returns reference start time of measurements.
    get_delta_time()
       Returns offset from the reference start time of measurement.
    get_instrument_settings()
       Returns instrument settings of measurement.
    get_gse_stimuli()
       Returns GSE stimuli parameters.
    get_exposure_time()
       Returns the exact pixel exposure time of the measurements.
    get_housekeeping_data()
       Returns housekeeping data of measurements.
    select(ic_id=None, *, msm_grp=None)
       Select a measurement as BAND%/ICID_<ic_id>_GROUP_%
    get_msm_attr(msm_dset, attr_name)
       Returns attribute of measurement dataset 'msm_dset'
    get_msm_data(msm_dset, fill_as_nan=True, frames=None, columns=None)
       Returns data of measurement dataset 'msm_dset'
    read_direct_msm(msm_dset, dest_sel=None, dest_dtype=None,
                    fill_as_nan=False)
       The faster implementation of class method 'get_msm_data'.

    Notes
    -----

    Examples
    --------
    """
    def __init__(self, ocm_product):
        """
        Initialize access to an OCAL Lx product

        Parameters
        ----------
        ocm_product :  string
           Full path to on-ground calibration measurement

        """
        if not Path(ocm_product).is_file():
            raise FileNotFoundError(f'{ocm_product} does not exist')

        # initialize class-attributes
        self.__msm_path = None
        self.band = None
        self.filename = ocm_product

        # open OCM product as HDF5 file
        self.fid = h5py.File(ocm_product, "r")

    def __repr__(self):
        class_name = type(self).__name__
        return f'{class_name}({self.filename!r})'

    def __iter__(self):
        for attr in sorted(self.__dict__):
            if not attr.startswith("__"):
                yield attr

    # def __del__(self):
    #    """
    #    called when the object is destroyed
    #    """
    #    self.close()

    def __enter__(self):
        """
        method called to initiate the context manager
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        method called when exiting the context manager
        """
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self):
        """
        Close resources
        """
        self.band = None
        if self.fid is not None:
            self.fid.close()
            self.fid = None

    # ---------- RETURN VERSION of the S/W ----------
    # ---------- Functions that work before MSM selection ----------
    def get_processor_version(self):
        """
        Returns version of the L01b processor
        """
        res = self.fid.attrs['processor_version']
        if isinstance(res, bytes):
            # pylint: disable=no-member
            res = res.decode('ascii')
        return res

    def get_coverage_time(self):
        """
        Returns start and end of the measurement coverage time
        """
        t_bgn = self.fid.attrs['time_coverage_start']
        if isinstance(t_bgn, bytes):
            # pylint: disable=no-member
            t_bgn = t_bgn.decode('ascii')

        t_end = self.fid.attrs['time_coverage_end']
        if isinstance(t_end, bytes):
            # pylint: disable=no-member
            t_end = t_end.decode('ascii')
        return (t_bgn, t_end)

    def get_attr(self, attr_name):
        """
        Obtain value of an HDF5 file attribute

        Parameters
        ----------
        attr_name : string
           name of the attribute
        """
        if attr_name in self.fid.attrs.keys():
            return self.fid.attrs[attr_name]

        return None

    # ---------- Functions that only work after MSM selection ----------
    def get_ref_time(self):
        """
        Returns reference start time of measurements
        """
        if not self.__msm_path:
            return {}

        grp = self.fid[f'BAND{self.band}']
        res = {}
        for msm in sorted(self.__msm_path):
            sgrp = grp[str(PurePosixPath(msm, 'GEODATA'))]
            res[msm] = datetime(2010, 1, 1, 0, 0, 0)
            res[msm] += timedelta(seconds=int(sgrp['time'][0]))

        return res

    def get_delta_time(self):
        """
        Returns offset from the reference start time of measurement
        """
        if not self.__msm_path:
            return {}

        grp = self.fid[f'BAND{self.band}']
        res = {}
        for msm in sorted(self.__msm_path):
            sgrp = grp[str(PurePosixPath(msm, 'GEODATA'))]
            res[msm] = sgrp['delta_time'][:].astype(int)

        return res

    def get_instrument_settings(self):
        """
        Returns instrument settings of measurement
        """
        if not self.__msm_path:
            return {}

        grp = self.fid[f'BAND{self.band}']
        res = {}
        for msm in sorted(self.__msm_path):
            sgrp = grp[str(PurePosixPath(msm, 'INSTRUMENT'))]
            res[msm] = np.squeeze(sgrp['instrument_settings'])

        return res

    def get_gse_stimuli(self):
        """
        Returns GSE stimuli parameters
        """
        if not self.__msm_path:
            return {}

        grp = self.fid[f'BAND{self.band}']
        res = {}
        for msm in sorted(self.__msm_path):
            sgrp = grp[str(PurePosixPath(msm, 'INSTRUMENT'))]
            res[msm] = np.squeeze(sgrp['gse_stimuli'])

        return res

    def get_exposure_time(self):
        """
        Returns the exact pixel exposure time of the measurements
        """
        if not self.__msm_path:
            return None

        grp = self.fid[f'BAND{self.band}']
        msm = self.__msm_path[0]  # all measurement sets have the same ICID
        sgrp = grp[str(PurePosixPath(msm, 'INSTRUMENT'))]
        instr = np.squeeze(sgrp['instrument_settings'])

        if int(self.band) > 6:
            return 1.25e-6 * (65540 - instr['int_delay'] + instr['int_hold'])

        return instr['exposure_time']

    def get_housekeeping_data(self):
        """
        Returns housekeeping data of measurements
        """
        if not self.__msm_path:
            return {}

        grp = self.fid[f'BAND{self.band}']
        res = {}
        for msm in sorted(self.__msm_path):
            sgrp = grp[str(PurePosixPath(msm, 'INSTRUMENT'))]
            res[msm] = np.squeeze(sgrp['housekeeping_data'])

        return res

    # -------------------------
    def select(self, ic_id=None, *, msm_grp=None):
        """
        Select a measurement as BAND%/ICID_<ic_id>_GROUP_%

        Parameters
        ----------
        ic_id  :  integer
          used as "BAND%/ICID_{}_GROUP_%".format(ic_id)
        msm_grp : string
          select measurement group with name msm_grp

        All measurements groups are shown when ic_id and msm_grp are None

        Returns
        -------
        out  :  scalar
           Number of measurements found

        Updated object attributes:
          - bands               : available spectral bands
        """
        self.band = ''
        self.__msm_path = []
        for ii in '87654321':
            if f'BAND{ii}' in self.fid:
                self.band = ii
                break

        if self.band:
            gid = self.fid[f'BAND{self.band}']
            if msm_grp is not None and msm_grp in gid:
                self.__msm_path = [msm_grp]
            elif ic_id is None:
                grp_name = 'ICID_'
                for kk in gid:
                    if kk.startswith(grp_name):
                        print(kk)
            else:
                grp_name = f'ICID_{ic_id:05}_GROUP'
                self.__msm_path = [s for s in gid if s.startswith(grp_name)]

        return len(self.__msm_path)

    # -------------------------
    def get_msm_attr(self, msm_dset, attr_name):
        """
        Returns attribute of measurement dataset "msm_dset"

        Parameters
        ----------
        msm_dset    :  string
            name of measurement dataset
        attr_name : string
            name of the attribute

        Returns
        -------
        out   :   scalar or numpy array
           value of attribute "attr_name"

        """
        if not self.__msm_path:
            return ''

        grp = self.fid[f'BAND{self.band}']
        for msm_path in self.__msm_path:
            ds_path = str(PurePosixPath(msm_path, 'OBSERVATIONS', msm_dset))

            if attr_name in grp[ds_path].attrs.keys():
                attr = grp[ds_path].attrs[attr_name]
                if isinstance(attr, bytes):
                    return attr.decode('ascii')

                return attr

        return None

    # -------------------------
    def get_msm_data(self, msm_dset, fill_as_nan=True,
                     frames=None, columns=None):
        """
        Returns data of measurement dataset "msm_dset"

        Parameters
        ----------
        msm_dset    :  string
            name of measurement dataset
            if msm_dset is None then show names of available datasets

        columns    : [i, j]
            Slice data on fastest axis (columns) as, from index 'i' to 'j'

        frames    : [i, j]
            Slice data on slowest axis (time) as, from index 'i' to 'j'

        fill_as_nan :  boolean
            replace (float) FillValues with Nan's

        Returns
        -------
        out   :   dictionary
           Python dictionary with names of msm_groups as keys
        """
        fillvalue = float.fromhex('0x1.ep+122')

        if not self.__msm_path:
            return {}

        # show HDF5 dataset names and return
        grp = self.fid[f'BAND{self.band}']
        if msm_dset is None:
            ds_path = str(PurePosixPath(self.__msm_path[0], 'OBSERVATIONS'))
            for kk in grp[ds_path]:
                print(kk)
            return {}

        # skip row257 from the SWIR detector
        rows = None
        if self.band in ('7', '8'):
            rows = [0, -1]

        # combine data of all measurement groups in dictionary
        res = {}
        for msm_grp in sorted(self.__msm_path):
            dset = grp[str(PurePosixPath(msm_grp, 'OBSERVATIONS', msm_dset))]
            data_sel = ()
            for ii in range(dset.ndim):
                dim_name = PurePosixPath(dset.dims[ii][0].name).name
                if dim_name == 'msmt_time':
                    if frames is None:
                        data_sel += (slice(None),)
                    else:
                        data_sel += (slice(*frames),)
                elif dim_name == 'row':
                    if rows is None:
                        data_sel += (slice(None),)
                    else:
                        data_sel += (slice(*rows),)
                elif dim_name == 'column':
                    if columns is None:
                        data_sel += (slice(None),)
                    else:
                        data_sel += (slice(*columns),)
                else:
                    raise ValueError

            # read data
            if dset.dtype == np.float32:
                data = np.squeeze(dset.astype(float)[data_sel])
            else:
                data = np.squeeze(dset[data_sel])

            if fill_as_nan and dset.attrs['_FillValue'] == fillvalue:
                data[(data == fillvalue)] = np.nan

            # add data to dictionary
            res[msm_grp] = data

        return res

    # -------------------------
    def read_direct_msm(self, msm_dset, dest_sel=None,
                        dest_dtype=None, fill_as_nan=False):
        """
        The faster implementation of get_msm_data()

        Parameters
        ----------
        msm_dset    :  string
            Name of measurement dataset
        dest_sel    :  numpy slice
            Selection must be the output of numpy.s_[<args>].
        dest_dtype  :  numpy dtype
            Perform type conversion
        fill_as_nan :  boolean
            Replace (float) FillValues with Nan's, when True

        Returns
        -------
        out   :   dictionary
           Python dictionary with names of msm_groups as keys
        """
        fillvalue = float.fromhex('0x1.ep+122')

        if not self.__msm_path:
            return None

        if dest_sel is None:
            dest_sel = np.s_[...]

        # combine data of all measurement groups in dictionary
        res = {}
        for msm_grp in sorted(self.__msm_path):
            dset = self.fid[str(PurePosixPath(
                f'BAND{self.band}', msm_grp,
                'OBSERVATIONS', msm_dset))]

            if dest_dtype is None:
                buff = dset[dest_sel]
            else:
                buff = dset.astype(dest_dtype)[dest_sel]

            if fill_as_nan:
                if dset.attrs['_FillValue'] == fillvalue:
                    buff[(buff == fillvalue)] = np.nan

            # add data to dictionary
            res[msm_grp] = buff

        return res



"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

Implementation of the Relative Least-Squares regression (RLS).

The RLS regression is used to find the linear dependence y(x) = c0 + c1 * x
that best describes the data before and after correction, using absolute
residuals y_i - (c0 + c1 * x_i) divided by the expected signals c1 * x_i in
the least-squares sum. Offset c0 has an arbitrary size and should not affect
the fit result. Weight factors are determined such to effectively spread the
data points evenly over the whole range of x, making the result less
sensitive to the actual spacing between the data points.

Copyright (c) 2019-2022 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import numpy as np
from numpy import ma


def calc_ma_weights(xdata: np.ndarray, masked: np.ndarray) -> ma.MaskedArray:
    """
    Generate weight factor per pixel

    Notes
    -----
    It might be that np.apply_along_axis() is slightly faster, however, the
    for-loop 'row in buff' is also very efficient when using the specially
    designed MaskedArray 'buff' which we eventually use to store the weight
    values per pixel.
    """
    buff = ma.array(np.repeat([xdata], masked.shape[0], axis=0), mask=masked)
    for row in buff:
        valid = ma.compressed(row)
        if len(valid) < 2:
            continue
        wght = np.concatenate(([2 * (valid[1] - valid[0])],
                               valid[2:] - valid[0:-2],
                               [2 * (valid[-1] - valid[-2])]))
        row[~row.mask] = wght

    return buff


def rls_fit(xdata: np.ndarray, ydata) -> tuple:
    """
    Perform RLS regression finding linear dependence y(x) = c0 + c1 * x

    Parameters
    ----------
    xdata :  ndarray, shape (M,)
       X-coordinates of the M sample points (xdata[i], ydata[..., i])
       The array values have to be positive and increasing
    ydata :  MaskedArray or ndarray, shape (..., M)
       Y-coordinates of the sample points

    Returns
    -------
    c0, c1, std_c0, std_c1 : tuple of ndarrays
       coefficients of the linear dependence and their standard deviations

    Raises
    ------
    RuntimeError
       if M < 2: too few points for a fit
       if M_xx != M_yy: number of samples not equal for xdata, ydata
       if ydata.shape[:-1] != samples_not_saturated.shape:
          arrays ydata and mask do not have equal shapes

    Notes
    -----
    Calling a rls-function with MaskedArrays is much slower than with
    plain ndarrays.

    The coefficients are set to NaN when the number of samples are less than 2.

    The standard deviations can only be calculated when the number of samples
    are larger than two, else the standard deviations are equal to zero.
    """
    # pylint: disable=too-many-locals
    if xdata.size < 2:
        raise RuntimeError('too few sample points for a fit')
    if xdata.size != ydata.shape[-1]:
        raise RuntimeError('number of samples not equal for xdata, ydata')

    # perform all computations on 2 dimensional arrays
    img_shape = ydata.shape[:-1]
    yy1 = ydata.reshape(-1, xdata.size)

    # calculate weights
    if ma.isMaskedArray(ydata):
        wghts = calc_ma_weights(xdata, ma.getmaskarray(yy1))
    else:
        buff = np.concatenate(([2 * (xdata[1] - xdata[0])],
                               xdata[2:] - xdata[0:-2],
                               [2 * (xdata[-1] - xdata[-2])]))
        wghts = np.repeat([buff], yy1.shape[0], axis=0)
    wx1 = wghts / xdata
    wx2 = wghts / xdata ** 2   # is wx1 / xdata faster? 

    # calculate the Q elements
    q00 = wghts.sum(axis=1)
    q01 = wx1.sum(axis=1)
    q02 = wx2.sum(axis=1)

    q11 = (wx1 * yy1).sum(axis=1)
    q12 = (wx2 * yy1).sum(axis=1)
    q22 = (wx2 * yy1 ** 2).sum(axis=1)

    # calculate the Z elements
    zz1 = q00 * q02 - q01 ** 2
    zz2 = q00 * q12 - q01 * q11
    zz3 = q02 * q11 - q01 * q12

    # calculate fit parameters and their uncertainties
    num = yy1.count(axis=1) if ma.isMaskedArray(ydata) else len(xdata)
    cc0 = zz2 / zz1
    cc1 = zz3 / zz1
    if ma.isMaskedArray(ydata):
        chi2 = ma.abs(q22 - q12 * cc0 - q11 * cc1) / np.clip(num - 2, 1, None)
        chi2[num <= 2] = 0
        sc0 = ma.sqrt(q00 * chi2 / zz1)
        sc1 = ma.sqrt(q02 * chi2 / zz1)

        return (cc0.reshape(img_shape).filled(np.nan),
                cc1.reshape(img_shape).filled(np.nan),
                sc0.reshape(img_shape).filled(np.nan),
                sc1.reshape(img_shape).filled(np.nan))

    # using only non-MaskedArray functions
    chi2 = np.abs(q22 - q12 * cc0 - q11 * cc1) / np.clip(num - 2, 1, None)
    chi2[num <= 2] = 0
    sc0 = np.sqrt(q00 * chi2 / zz1)
    sc1 = np.sqrt(q02 * chi2 / zz1)

    return (cc0.reshape(img_shape), cc1.reshape(img_shape),
            sc0.reshape(img_shape), sc1.reshape(img_shape))




def rls_fit0(xdata: np.ndarray, ydata) -> tuple:
    """
    Perform RLS regression finding linear dependence y(x) = c1 * x

    Parameters
    ----------
    xdata :  ndarray, shape (M,)
       X-coordinates of the M sample points (xdata[i], ydata[..., i])
       The array values have to be positive and increasing
    ydata :  MaskedArray or ndarray, shape (..., M)
       Y-coordinates of the sample points

    Returns
    -------
    c1, std_c1 : tuple of ndarrays
       coefficients of the linear dependence and their standard deviations

    Raises
    ------
    RuntimeError
       if M < 2: too few points for a fit
       if M_xx != M_yy: number of samples not equal for xdata, ydata
       if ydata.shape[:-1] != samples_not_saturated.shape:
          arrays ydata and mask do not have equal shapes

    Notes
    -----
    The coefficients are set to NaN when the number of samples are less than 2.

    The standard deviations can only be calculated when the number of samples
    are larger than two, else the standard deviations are equal to zero.
    """
    if xdata.size < 2:
        raise RuntimeError('too few points for a fit')
    if xdata.size != ydata.shape[-1]:
        raise RuntimeError('number of samples not equal for xdata, ydata')

    # perform all computations on 2 dimensional arrays
    img_shape = ydata.shape[:-1]
    yy1 = ydata.reshape(-1, xdata.size)

    # calculate weights
    if ma.isMaskedArray(ydata):
        wghts = calc_ma_weights(xdata, ma.getmaskarray(yy1))
    else:
        buff = np.concatenate(([2 * (xdata[1] - xdata[0])],
                               xdata[2:] - xdata[0:-2],
                               [2 * (xdata[-1] - xdata[-2])]))
        wghts = np.repeat([buff], yy1.shape[0], axis=0)
    wx1 = wghts / xdata
    wx2 = wghts / xdata ** 2

    # calculate the Q elements
    q00 = wghts.sum(axis=1)
    q11 = (wx1 * yy1).sum(axis=1)
    q22 = (wx2 * yy1 ** 2).sum(axis=1)

    # calculate fit parameter and its variance
    num = yy1.count(axis=1) if ma.isMaskedArray(ydata) else len(xdata)
    cc1 = q11 / q00
    if ma.isMaskedArray(ydata):
        cc1[num < 1] = ma.masked
        chi2 = ma.abs(q22 - q00 * cc1 ** 2) / np.clip(num - 1, 1, None)
        chi2[num <= 1] = ma.masked
        sc1 = ma.sqrt(chi2 / q00)
        return (cc1.reshape(img_shape).filled(np.nan),
                sc1.reshape(img_shape).filled(np.nan))

    # using only non-MaskedArray functions
    cc1[num < 1] = np.nan
    chi2 = np.abs(q22 - q00 * cc1 ** 2) / np.clip(num - 1, 1, None)
    chi2[num <= 1] = np.nan
    sc1 = np.sqrt(chi2 / q00)
    return (cc1.reshape(img_shape), sc1.reshape(img_shape))



"""
This file is part of pyS5p

https://github.com/rmvanhees/pys5p.git

The classes L1Bio, L1BioIRR, L1BioRAD and L1BioENG provide read/write access to
offline level 1b products, resp. calibration, irradiance, radiance
and engineering.

Copyright (c) 2017-2022 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from datetime import datetime, timedelta
from pathlib import Path, PurePosixPath
from setuptools_scm import get_version

import h5py
import numpy as np

from moniplot.biweight import biweight

from .swir_texp import swir_exp_time

# - global parameters ------------------------------


# - local functions --------------------------------
def pad_rows(arr1, arr2):
    """
    Pad the array with the least numer of rows with NaN's
    """
    if arr2.ndim == 1:
        pass
    elif arr2.ndim == 2:
        if arr1.shape[0] < arr2.shape[0]:
            buff = arr1.copy()
            arr1 = np.full(arr2.shape, np.nan, dtype=arr2.dtype)
            arr1[0:buff.shape[0], :] = buff
        elif arr1.shape[0] > arr2.shape[0]:
            buff = arr2.copy()
            arr2 = np.full(arr1.shape, np.nan, dtype=arr2.dtype)
            arr2[0:buff.shape[0], :] = buff
    else:
        if arr1.shape[1] < arr2.shape[1]:
            buff = arr1.copy()
            arr1 = np.full(arr2.shape, np.nan, dtype=arr2.dtype)
            arr1[:, 0:buff.shape[1], :] = buff
        elif arr1.shape[1] > arr2.shape[1]:
            buff = arr2.copy()
            arr2 = np.full(arr1.shape, np.nan, dtype=arr2.dtype)
            arr2[:, 0:buff.shape[1], :] = buff

    return (arr1, arr2)


# - class definition -------------------------------
class L1Bio:
    """
    class with methods to access Tropomi L1B calibration products

    The L1b calibration products are available for UVN (band 1-6)
    and SWIR (band 7-8).

    Attributes
    ----------
    fid : h5py.File
    filename : string
    bands : string

    Methods
    -------
    close()
       Close recources.
    get_attr(attr_name)
       Obtain value of an HDF5 file attribute.
    get_orbit()
       Returns absolute orbit number.
    get_processor_version()
       Returns version of the L01b processor used to generate this product.
    get_coverage_time()
       Returns start and end of the measurement coverage time.
    get_creation_time()
       Returns datetime when the L1b product was created.
    select(msm_type=None)
       Select a calibration measurement as <processing class>_<ic_id>.
    sequence(band=None)
       Returns sequence number for each unique measurement based on ICID
       and delta_time.
    get_ref_time(band=None)
       Returns reference start time of measurements.
    get_delta_time(band=None)
       Returns offset from the reference start time of measurement.
    get_instrument_settings(band=None)
       Returns instrument settings of measurement.
    get_exposure_time(band=None)
       Returns pixel exposure time of the measurements, which is calculated
       from the parameters 'int_delay' and 'int_hold' for SWIR.
    get_housekeeping_data(band=None)
       Returns housekeeping data of measurements.
    get_geo_data(geo_dset=None, band=None)
       Returns data of selected datasets from the GEODATA group.
    get_msm_attr(msm_dset, attr_name, band=None)
       Returns value attribute of measurement dataset "msm_dset".
    get_msm_data(msm_dset, band=None, fill_as_nan=False, msm_to_row=None)
       Reads data from dataset "msm_dset".
    set_msm_data(msm_dset, new_data)
       Replace data of dataset "msm_dset" with new_data.

    Notes
    -----

    Examples
    --------
    """
    band_groups = ('/BAND%_CALIBRATION', '/BAND%_IRRADIANCE',
                   '/BAND%_RADIANCE')
    geo_dset = 'satellite_latitude,satellite_longitude'
    msm_type = None

    def __init__(self, l1b_product, readwrite=False, verbose=False):
        """
        Initialize access to a Tropomi offline L1b product
        """
        # open L1b product as HDF5 file
        if not Path(l1b_product).is_file():
            raise FileNotFoundError(f'{l1b_product} does not exist')

        # initialize private class-attributes
        self.__rw = readwrite
        self.__verbose = verbose
        self.__msm_path = None
        self.__patched_msm = []
        self.filename = l1b_product
        self.bands = ''

        if readwrite:
            self.fid = h5py.File(l1b_product, "r+")
        else:
            self.fid = h5py.File(l1b_product, "r")

    def __repr__(self):
        class_name = type(self).__name__
        return f'{class_name}({self.filename!r}, readwrite={self.__rw!r})'

    def __iter__(self):
        for attr in sorted(self.__dict__):
            if not attr.startswith("__"):
                yield attr

    def __enter__(self):
        """
        method called to initiate the context manager
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        method called when exiting the context manager
        """
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self):
        """
        Close resources.

        Notes
        -----
        Before closing the product, we make sure that the output product
        describes what has been altered by the S/W. To keep any change
        traceable.

        In case the L1b product is altered, the attributes listed below are
        added to the group: "/METADATA/SRON_METADATA":
             - dateStamp ('now')
             - Git-version of S/W
             - list of patched datasets
             - auxiliary datasets used by patch-routines
        """
        if self.fid is None:
            return

        if self.__patched_msm:
            # pylint: disable=no-member
            sgrp = self.fid.require_group("/METADATA/SRON_METADATA")
            sgrp.attrs['dateStamp'] = datetime.utcnow().isoformat()
            sgrp.attrs['git_tag'] = get_version(root='..',
                                                relative_to=__file__)
            if 'patched_datasets' not in sgrp:
                dtype = h5py.special_dtype(vlen=str)
                dset = sgrp.create_dataset('patched_datasets',
                                           (len(self.__patched_msm),),
                                           maxshape=(None,), dtype=dtype)
                dset[:] = np.asarray(self.__patched_msm)
            else:
                dset = sgrp['patched_datasets']
                dset.resize(dset.shape[0] + len(self.__patched_msm), axis=0)
                dset[dset.shape[0]-1:] = np.asarray(self.__patched_msm)

        self.fid.close()
        self.fid = None

    # ---------- PUBLIC FUNCTIONS ----------
    def get_attr(self, attr_name):
        """
        Obtain value of an HDF5 file attribute

        Parameters
        ----------
        attr_name :  string
           Name of the attribute
        """
        if attr_name not in self.fid.attrs.keys():
            return None

        attr = self.fid.attrs[attr_name]
        if attr.shape is None:
            return None

        return attr

    def get_orbit(self):
        """
        Returns absolute orbit number
        """
        res = self.get_attr('orbit')
        if res is None:
            return None

        return int(res)

    def get_processor_version(self):
        """
        Returns version of the L01b processor
        """
        attr = self.get_attr('processor_version')
        if attr is None:
            return None

        # pylint: disable=no-member
        return attr.decode('ascii')

    def get_coverage_time(self):
        """
        Returns start and end of the measurement coverage time
        """
        attr_start = self.get_attr('time_coverage_start')
        if attr_start is None:
            return None

        attr_end = self.get_attr('time_coverage_end')
        if attr_end is None:
            return None

        # pylint: disable=no-member
        return (attr_start.decode('ascii'),
                attr_end.decode('ascii'))

    def get_creation_time(self):
        """
        Returns datetime when the L1b product was created
        """
        grp = self.fid['/METADATA/ESA_METADATA/earth_explorer_header']
        dset = grp['fixed_header/source']
        if 'Creation_Date' in self.fid.attrs.keys():
            attr = dset.attrs['Creation_Date']
            if isinstance(attr, bytes):
                return attr.decode('ascii')

            return attr

        return None

    def select(self, msm_type=None):
        """
        Select a calibration measurement as <processing class>_<ic_id>

        Parameters
        ----------
        msm_type :  string
          Name of calibration measurement group as <processing class>_<ic_id>

        Returns
        -------
        out  :   string
           String with spectral bands found in product

        Updated object attributes:
         - bands               : available spectral bands
        """
        if msm_type is None:
            if self.msm_type is None:
                raise ValueError('parameter msm_type is not defined')
            msm_type = self.msm_type

        self.bands = ''
        for name in self.band_groups:
            for ii in '12345678':
                grp_path = PurePosixPath(name.replace('%', ii), msm_type)
                if str(grp_path) in self.fid:
                    if self.__verbose:
                        print('*** INFO: found: ', grp_path)
                    self.bands += ii

            if self.bands:
                self.__msm_path = str(
                    PurePosixPath(name, msm_type))
                break

        return self.bands

    def sequence(self, band=None):
        """
        Returns sequence number for each unique measurement based on ICID
          and delta_time

        Parameters
        ----------
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product
            Default is 'None' which returns the first available band

        Returns
        -------
        out  :  array-like
          Numpy rec-array with sequence number, ICID and delta-time
        """
        if self.__msm_path is None:
            return None

        if band is None or len(band) > 1:
            band = self.bands[0]

        msm_path = self.__msm_path.replace('%', band)
        grp = self.fid[str(PurePosixPath(msm_path, 'INSTRUMENT'))]

        icid_list = np.squeeze(grp['instrument_configuration']['ic_id'])
        master_cycle = grp['instrument_settings']['master_cycle_period_us'][0]
        master_cycle /= 1000
        grp = self.fid[str(PurePosixPath(msm_path, 'OBSERVATIONS'))]
        delta_time = np.squeeze(grp['delta_time'])

        # define result as numpy array
        length = delta_time.size
        res = np.empty((length,), dtype=[('sequence', 'u2'),
                                         ('icid', 'u2'),
                                         ('delta_time', 'u4'),
                                         ('index', 'u4')])
        res['sequence'] = [0]
        res['icid'] = icid_list
        res['delta_time'] = delta_time
        res['index'] = np.arange(length, dtype=np.uint32)
        if length == 1:
            return res

        # determine sequence number
        buff_icid = np.concatenate(([icid_list[0]-10], icid_list,
                                    [icid_list[-1]+10]))
        dt_thres = 10 * master_cycle
        buff_time = np.concatenate(([delta_time[0] - 10 * dt_thres], delta_time,
                                    [delta_time[-1] + 10 * dt_thres]))

        indx = (((buff_time[1:] - buff_time[0:-1]) > dt_thres)
                | ((buff_icid[1:] - buff_icid[0:-1]) != 0)).nonzero()[0]
        for ii in range(len(indx)-1):
            res['sequence'][indx[ii]:indx[ii+1]] = ii

        return res

    def get_ref_time(self, band=None):
        """
        Returns reference start time of measurements

        Parameters
        ----------
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product.
            Default is 'None' which returns the first available band
        """
        if self.__msm_path is None:
            return None

        if band is None:
            band = self.bands[0]

        msm_path = self.__msm_path.replace('%', band)
        grp = self.fid[str(PurePosixPath(msm_path, 'OBSERVATIONS'))]

        return datetime(2010, 1, 1, 0, 0, 0) \
            + timedelta(seconds=int(grp['time'][0]))

    def get_delta_time(self, band=None):
        """
        Returns offset from the reference start time of measurement

        Parameters
        ----------
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product.
            Default is 'None' which returns the first available band
        """
        if self.__msm_path is None:
            return None

        if band is None:
            band = self.bands[0]

        msm_path = self.__msm_path.replace('%', band)
        grp = self.fid[str(PurePosixPath(msm_path, 'OBSERVATIONS'))]

        return grp['delta_time'][0, :].astype(int)

    def get_instrument_settings(self, band=None):
        """
        Returns instrument settings of measurement

        Parameters
        ----------
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product.
            Default is 'None' which returns the first available band
        """
        if self.__msm_path is None:
            return None

        if band is None:
            band = self.bands[0]

        msm_path = self.__msm_path.replace('%', band)
        #
        # Due to a bug in python module h5py (v2.6.0), it fails to read
        # the UVN instrument settings directy, with exception:
        #    KeyError: 'Unable to open object (Component not found)'.
        # This is my workaround
        #
        grp = self.fid[str(PurePosixPath(msm_path, 'INSTRUMENT'))]
        instr = np.empty(grp['instrument_settings'].shape,
                         dtype=grp['instrument_settings'].dtype)
        grp['instrument_settings'].read_direct(instr)
        # for name in grp['instrument_settings'].dtype.names:
        #     instr[name][:] = grp['instrument_settings'][name]

        return instr

    def get_exposure_time(self, band=None):
        """
        Returns pixel exposure time of the measurements, which is calculated
        from the parameters 'int_delay' and 'int_hold' for SWIR.

        Parameters
        ----------
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product
            Default is 'None' which returns the first available band
        """
        if band is None:
            band = self.bands[0]

        instr_arr = self.get_instrument_settings(band)

        # calculate exact exposure time
        if int(band) < 7:
            return [instr['exposure_time'] for instr in instr_arr]

        return [swir_exp_time(instr['int_delay'], instr['int_hold'])
                for instr in instr_arr]

    def get_housekeeping_data(self, band=None):
        """
        Returns housekeeping data of measurements

        Parameters
        ----------
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product
            Default is 'None' which returns the first available band
        """
        if self.__msm_path is None:
            return None

        if band is None:
            band = self.bands[0]

        msm_path = self.__msm_path.replace('%', band)
        grp = self.fid[str(PurePosixPath(msm_path, 'INSTRUMENT'))]

        return np.squeeze(grp['housekeeping_data'])

    def get_geo_data(self, geo_dset=None, band=None):
        """
        Returns data of selected datasets from the GEODATA group

        Parameters
        ----------
        geo_dset  :  string
            Name(s) of datasets in the GEODATA group, comma separated
            Default is 'satellite_latitude,satellite_longitude'
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product
            Default is 'None' which returns the first available band

        Returns
        -------
        out   :   dict of numpy
           Compound array with data of selected datasets from the GEODATA group
        """
        if self.__msm_path is None:
            return None

        if geo_dset is None:
            geo_dset = self.geo_dset

        if band is None:
            band = self.bands[0]

        msm_path = self.__msm_path.replace('%', band)
        grp = self.fid[str(PurePosixPath(msm_path, 'GEODATA'))]

        res = {}
        for name in geo_dset.split(','):
            res[name] = grp[name][0, ...]

        return res

    def get_msm_attr(self, msm_dset, attr_name, band=None):
        """
        Returns value attribute of measurement dataset "msm_dset"

        Parameters
        ----------
        attr_name : string
            Name of the attribute
        msm_dset  :  string
            Name of measurement dataset
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product
            Default is 'None' which returns the first available band

        Returns
        -------
        out   :   scalar or numpy array
            Value of attribute "attr_name"
        """
        if self.__msm_path is None:
            return None

        if band is None:
            band = self.bands[0]

        msm_path = self.__msm_path.replace('%', band)
        ds_path = str(PurePosixPath(msm_path, 'OBSERVATIONS', msm_dset))
        if attr_name in self.fid[ds_path].attrs.keys():
            attr = self.fid[ds_path].attrs[attr_name]
            if isinstance(attr, bytes):
                return attr.decode('ascii')

            return attr

        return None

    def get_msm_data(self, msm_dset, band=None,
                     fill_as_nan=False, msm_to_row=None):
        """
        Reads data from dataset "msm_dset"

        Parameters
        ----------
        msm_dset :  string
            Name of measurement dataset.
        band      :  None or {'1', '2', '3', ..., '8'}
            Select one of the band present in the product
            Default is 'None' which returns
                  both bands (Calibration, Irradiance)
                  or one band (Radiance)
        fill_as_nan :  boolean
            Set data values equal (KNMI) FillValue to NaN
        msm_to_row : boolean
            Combine two bands using padding if necessary

        Returns
        -------
        out   :  values read from or written to dataset "msm_dset"
        """
        fillvalue = float.fromhex('0x1.ep+122')

        if self.__msm_path is None:
            return None

        if band is None:
            band = self.bands
        elif not isinstance(band, str):
            raise TypeError('band must be a string')
        elif band not in self.bands:
            raise ValueError('band not found in product')

        if len(band) == 2 and msm_to_row is None:
            msm_to_row = 'padding'

        data = ()
        for ii in band:
            msm_path = self.__msm_path.replace('%', ii)
            ds_path = str(PurePosixPath(msm_path, 'OBSERVATIONS', msm_dset))
            dset = self.fid[ds_path]

            if fill_as_nan and dset.attrs['_FillValue'] == fillvalue:
                buff = np.squeeze(dset)
                buff[(buff == fillvalue)] = np.nan
                data += (buff,)
            else:
                data += (np.squeeze(dset),)

        if len(band) == 1:
            return data[0]

        if msm_to_row == 'padding':
            data = pad_rows(data[0], data[1])

        return np.concatenate(data, axis=data[0].ndim-1)

    def set_msm_data(self, msm_dset, new_data):
        """
        Replace data of dataset "msm_dset" with new_data

        Parameters
        ----------
        msm_dset   :  string
            Name of measurement dataset.
        new_data :  array-like
            Data to be written with same dimensions as dataset "msm_dset"
        """
        if self.__msm_path is None:
            return

        # we will overwrite existing data, thus readwrite access is required
        if not self.__rw:
            raise PermissionError('read/write access required')

        # overwrite the data
        col = 0
        for ii in self.bands:
            msm_path = self.__msm_path.replace('%', ii)
            ds_path = str(PurePosixPath(msm_path, 'OBSERVATIONS', msm_dset))
            dset = self.fid[ds_path]

            dims = dset.shape
            dset[0, ...] = new_data[..., col:col+dims[-1]]
            col += dims[-1]

            # update patch logging
            self.__patched_msm.append(ds_path)


# --------------------------------------------------
class L1BioIRR(L1Bio):
    """
    class with methods to access Tropomi L1B irradiance products
    """
    band_groups = ('/BAND%_IRRADIANCE',)
    geo_dset = 'earth_sun_distance'
    msm_type = 'STANDARD_MODE'


# --------------------------------------------------
class L1BioRAD(L1Bio):
    """
    class with function to access Tropomi L1B radiance products
    """
    band_groups = ('/BAND%_RADIANCE',)
    geo_dset = 'latitude,longitude'
    msm_type = 'STANDARD_MODE'


# --------------------------------------------------
class L1BioENG:
    """
    class with methods to access Tropomi offline L1b engineering products

    Attributes
    ----------
    fid : HDF5 file object
    filename : string

    Methods
    -------
    close()
       Close recources.
    get_attr(attr_name)
       Obtain value of an HDF5 file attribute.
    get_orbit()
       Returns absolute orbit number.
    get_processor_version()
       Returns version of the L01b processor used to generate this product.
    get_coverage_time()
       Returns start and end of the measurement coverage time.
    get_creation_time()
       Returns datetime when the L1b product was created.
    get_ref_time()
       Returns reference start time of measurements.
    get_delta_time()
       Returns offset from the reference start time of measurement.
    get_msmtset()
       Returns L1B_ENG_DB/SATELLITE_INFO/satellite_pos.
    get_msmtset_db()
       Returns compressed msmtset from L1B_ENG_DB/MSMTSET/msmtset.
    get_swir_hk_db(stats=None, fill_as_nan=False)
       Returns the most important SWIR house keeping parameters.

    Notes
    -----
    The L1b engineering products are available for UVN (band 1-6)
    and SWIR (band 7-8).

    Examples
    --------
    """
    def __init__(self, l1b_product):
        """
        Initialize access to a Tropomi offline L1b product
        """
        # open L1b product as HDF5 file
        if not Path(l1b_product).is_file():
            raise FileNotFoundError(f'{l1b_product} does not exist')

        # initialize private class-attributes
        self.filename = l1b_product
        self.fid = h5py.File(l1b_product, "r")

    def __repr__(self):
        class_name = type(self).__name__
        return f'{class_name}({self.filename!r})'

    def __iter__(self):
        for attr in sorted(self.__dict__):
            if not attr.startswith("__"):
                yield attr

    def __enter__(self):
        """
        method called to initiate the context manager
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        method called when exiting the context manager
        """
        self.close()
        return False  # any exception is raised by the with statement.

    def close(self):
        """
        close access to product
        """
        if self.fid is None:
            return

        self.fid.close()
        self.fid = None

    # ---------- PUBLIC FUNCTIONS ----------
    def get_attr(self, attr_name):
        """
        Obtain value of an HDF5 file attribute

        Parameters
        ----------
        attr_name :  string
           Name of the attribute
        """
        if attr_name not in self.fid.attrs.keys():
            return None

        attr = self.fid.attrs[attr_name]
        if attr.shape is None:
            return None

        return attr

    def get_orbit(self):
        """
        Returns absolute orbit number
        """
        res = self.get_attr('orbit')
        if res is None:
            return None

        return int(res)

    def get_processor_version(self):
        """
        Returns version of the L01b processor
        """
        attr = self.get_attr('processor_version')
        if attr is None:
            return None

        # pylint: disable=no-member
        return attr.decode('ascii')

    def get_coverage_time(self):
        """
        Returns start and end of the measurement coverage time
        """
        attr_start = self.get_attr('time_coverage_start')
        if attr_start is None:
            return None

        attr_end = self.get_attr('time_coverage_end')
        if attr_end is None:
            return None

        # pylint: disable=no-member
        return (attr_start.decode('ascii'),
                attr_end.decode('ascii'))

    def get_creation_time(self):
        """
        Returns datetime when the L1b product was created
        """
        grp = self.fid['/METADATA/ESA_METADATA/earth_explorer_header']
        dset = grp['fixed_header/source']
        if 'Creation_Date' in self.fid.attrs.keys():
            attr = dset.attrs['Creation_Date']
            if isinstance(attr, bytes):
                return attr.decode('ascii')

            return attr

        return None

    def get_ref_time(self):
        """
        Returns reference start time of measurements
        """
        return self.fid['reference_time'][0].astype(int)

    def get_delta_time(self):
        """
        Returns offset from the reference start time of measurement
        """
        return self.fid['/MSMTSET/msmtset']['delta_time'][:].astype(int)

    def get_msmtset(self):
        """
        Returns L1B_ENG_DB/SATELLITE_INFO/satellite_pos
        """
        return self.fid['/SATELLITE_INFO/satellite_pos'][:]

    def get_msmtset_db(self):
        """
        Returns compressed msmtset from L1B_ENG_DB/MSMTSET/msmtset

        Notes
        -----
        This function is used to fill the SQLite product databases
        """
        dtype_msmt_db = np.dtype([('meta_id', np.int32),
                                  ('ic_id', np.uint16),
                                  ('ic_version', np.uint8),
                                  ('class', np.uint8),
                                  ('repeats', np.uint16),
                                  ('exp_per_mcp', np.uint16),
                                  ('exp_time_us', np.uint32),
                                  ('mcp_us', np.uint32),
                                  ('delta_time_start', np.int32),
                                  ('delta_time_end', np.int32)])

        # read full msmtset
        msmtset = self.fid['/MSMTSET/msmtset'][:]

        # get indices to start and end of every measurement (based in ICID)
        icid = msmtset['icid']
        indx = (np.diff(icid) != 0).nonzero()[0] + 1
        indx = np.insert(indx, 0, 0)
        indx = np.append(indx, -1)

        # compress data from msmtset
        msmt = np.zeros(indx.size-1, dtype=dtype_msmt_db)
        msmt['ic_id'][:] = msmtset['icid'][indx[0:-1]]
        msmt['ic_version'][:] = msmtset['icv'][indx[0:-1]]
        msmt['class'][:] = msmtset['class'][indx[0:-1]]
        msmt['delta_time_start'][:] = msmtset['delta_time'][indx[0:-1]]
        msmt['delta_time_end'][:] = msmtset['delta_time'][indx[1:]]

        # add SWIR timing information
        timing = self.fid['/DETECTOR4/timing'][:]
        msmt['mcp_us'][:] = timing['mcp_us'][indx[1:]-1]
        msmt['exp_time_us'][:] = timing['exp_time_us'][indx[1:]-1]
        msmt['exp_per_mcp'][:] = timing['exp_per_mcp'][indx[1:]-1]
        # duration per ICID execution in micro-seconds
        duration = 1000 * (msmt['delta_time_end'] - msmt['delta_time_start'])
        # duration can be zero
        mask = msmt['mcp_us'] > 0
        # divide duration by measurement period in micro-seconds
        msmt['repeats'][mask] = (duration[mask]
                                 / (msmt['mcp_us'][mask])).astype(np.uint16)

        return msmt

    def get_swir_hk_db(self, stats=None, fill_as_nan=False):
        """
        Returns the most important SWIR house keeping parameters

        Parameters
        ----------
        fill_as_nan :  boolean
            Replace (float) FillValues with Nan's, when True

        Notes
        -----
        This function is used to fill the SQLite product datbase and
        HDF5 monitoring database
        """
        dtype_hk_db = np.dtype([('detector_temp', np.float32),
                                ('grating_temp', np.float32),
                                ('imager_temp', np.float32),
                                ('obm_temp', np.float32),
                                ('calib_unit_temp', np.float32),
                                ('fee_inner_temp', np.float32),
                                ('fee_board_temp', np.float32),
                                ('fee_ref_volt_temp', np.float32),
                                ('fee_video_amp_temp', np.float32),
                                ('fee_video_adc_temp', np.float32),
                                ('detector_heater', np.float32),
                                ('obm_heater_cycle', np.float32),
                                ('fee_box_heater_cycle', np.float32),
                                ('obm_heater', np.float32),
                                ('fee_box_heater', np.float32)])

        num_eng_pkts = self.fid['nr_of_engdat_pkts'].size
        swir_hk = np.empty(num_eng_pkts, dtype=dtype_hk_db)

        hk_tbl = self.fid['/DETECTOR4/DETECTOR_HK/temperature_info'][:]
        swir_hk['detector_temp'] = hk_tbl['temp_det_ts2']
        swir_hk['fee_inner_temp'] = hk_tbl['temp_d1_box']
        swir_hk['fee_board_temp'] = hk_tbl['temp_d5_cold']
        swir_hk['fee_ref_volt_temp'] = hk_tbl['temp_a3_vref']
        swir_hk['fee_video_amp_temp'] = hk_tbl['temp_d6_vamp']
        swir_hk['fee_video_adc_temp'] = hk_tbl['temp_d4_vadc']

        hk_tbl = self.fid['/NOMINAL_HK/TEMPERATURES/hires_temperatures'][:]
        swir_hk['grating_temp'] = hk_tbl['hires_temp_1']

        hk_tbl = self.fid['/NOMINAL_HK/TEMPERATURES/instr_temperatures'][:]
        swir_hk['imager_temp'] = hk_tbl['instr_temp_29']
        swir_hk['obm_temp'] = hk_tbl['instr_temp_28']
        swir_hk['calib_unit_temp'] = hk_tbl['instr_temp_25']

        hk_tbl = self.fid['/DETECTOR4/DETECTOR_HK/heater_data'][:]
        swir_hk['detector_heater'] = hk_tbl['det_htr_curr']

        hk_tbl = self.fid['/NOMINAL_HK/HEATERS/heater_data'][:]
        swir_hk['obm_heater'] = hk_tbl['meas_cur_val_htr12']
        swir_hk['obm_heater_cycle'] = hk_tbl['last_pwm_val_htr12']
        swir_hk['fee_box_heater'] = hk_tbl['meas_cur_val_htr13']
        swir_hk['fee_box_heater_cycle'] = hk_tbl['last_pwm_val_htr13']

        # CHECK: works only when all elements of swir_hk are floats
        if fill_as_nan:
            for key in dtype_hk_db.names:
                swir_hk[key][swir_hk[key] == 999.] = np.nan

        if stats is None:
            return swir_hk

        if stats == 'median':
            hk_median = np.empty(1, dtype=dtype_hk_db)
            for key in dtype_hk_db.names:
                if np.all(np.isnan(swir_hk[key])):
                    hk_median[key][0] = np.nan
                elif np.nanmin(swir_hk[key]) == np.nanmax(swir_hk[key]):
                    hk_median[key][0] = swir_hk[key][0]
                else:
                    hk_median[key][0] = biweight(swir_hk[key])
            return hk_median

        if stats == 'range':
            hk_min = np.empty(1, dtype=dtype_hk_db)
            hk_max = np.empty(1, dtype=dtype_hk_db)
            for key in dtype_hk_db.names:
                if np.all(np.isnan(swir_hk[key])):
                    hk_min[key][0] = np.nan
                    hk_max[key][0] = np.nan
                elif np.nanmin(swir_hk[key]) == np.nanmax(swir_hk[key]):
                    hk_min[key][0] = swir_hk[key][0]
                    hk_max[key][0] = swir_hk[key][0]
                else:
                    hk_min[key][0] = np.nanmin(swir_hk[key])
                    hk_max[key][0] = np.nanmax(swir_hk[key])
            return (hk_min, hk_max)

        return None
